/*
		This program is free software; you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
		GNU General Public License for more details.

		You should have received a copy of the GNU General Public License
		along with this program; if not, write to the Free Software
		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

		The author can be contacted at:
		Donald Graft
		neuron2@attbi.com.
*/

#ifdef AVISYNTH25
#include "internal.h"
#else
#include "avisynth.h"
#endif
#include "windows.h"
#include "stdio.h"
#include "version.h"

extern void DrawString(PVideoFrame &dst, int x, int y, const char *s); 

class FieldDeinterlace: public GenericVideoFilter
{
	bool full, debug, blend, chroma, show;
	int threshold, dthreshold;
	const char *ovr;
	unsigned int *overrides, *overrides_p;
	PVideoFrame src, fmask, dmask, final;
	const unsigned char *srcp, *srcp_saved, *p, *n;
	unsigned char *fmaskp, *fmaskp_saved, *dmaskp, *dmaskp_saved, *finalp;
	int pitch, dpitch, w, h;
	bool MotionMask_YUY2(int frame, IScriptEnvironment* env);

public:
	FieldDeinterlace(PClip _child,
					bool _full, int _threshold, int _dthreshold, bool _blend, bool _chroma,
					const char * _ovr, bool _show, bool _debug,
					IScriptEnvironment* env) : GenericVideoFilter(_child), full(_full),
					threshold(_threshold), dthreshold(_dthreshold), blend(_blend), chroma(_chroma),
					ovr(_ovr), show(_show), debug(_debug)
	{
		FILE *f;
		int count = 0;
		char buf[80];
		char c, *d;
		unsigned int *p;

		if (!vi.IsYUY2())
			env->ThrowError("FieldDeinterlace: YUY2 data only (no RGB); use ConvertToYUY2");
		/* Load manual overrides file if there is one. */
		overrides = NULL;
		if (ovr && (f = fopen(ovr, "r")) != NULL)
		{
			while(fgets(buf, 80, f) != 0)
			{
				count++;
			}
			fclose(f);
			if ((f = fopen(ovr, "r")) != NULL)
			{
				p = overrides = (unsigned int *) malloc(3 * (count + 1) * sizeof(unsigned int));
				memset(overrides, 0xff, 3 * (count + 1) * sizeof(unsigned int));
				while(fgets(buf, 80, f) != 0)
				{
					d = buf;
					while (*d != ',' && *d != 0) d++;
					if (*d == 0)
					{
						sscanf(buf, "%d %c", p, &c);
						p++;
						*p = *(p-1);
						p++;
						*p = c;
						p++;
					}
					else
					{
						sscanf(buf, "%d,%d %c", p, p+1, &c);
						p += 2;
						*p = c;
						p++;
					}
				}
				fclose(f);
			}
		}
		if (debug)
		{
			char b[80];
			sprintf(b, "FieldDeinterlace %s by Donald Graft, Copyright 2002\n", VERSION);
			OutputDebugString(b);
		}
	}

	FieldDeinterlace::~FieldDeinterlace(void)
	{
		if (overrides != NULL) free(overrides);
	}

	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
};

bool FieldDeinterlace::MotionMask_YUY2(int frame, IScriptEnvironment* env)
{
	int x, y, thresh, dthresh;
	int *counts, box, boxy;
#ifdef DEINTERLACE_MMX_BUILD
	void asm_deinterlace(const unsigned char *srcp, const unsigned char *p,
								  const unsigned char *n, unsigned char *fmask,
								  unsigned char *dmask, int thresh, int dthresh, int row_size);
	void asm_deinterlace_chroma(const unsigned char *srcp, const unsigned char *p,
								  const unsigned char *n, unsigned char *fmask,
								  unsigned char *dmask, int thresh, int dthresh, int row_size);
#else
	int val, val2;
#endif
	const int CT = 15;
	int boxarraysize;
	char buf[80];

	thresh = threshold*threshold;
	dthresh = dthreshold*dthreshold;

	p = srcp - pitch;
	n = srcp + pitch;
	memset(fmaskp, 0, w);
	for (x = 0; x < w; x += 4) *(int *)(dmaskp+x) = 0x00ff00ff;

	for (y = 1; y < h - 1; y++)
	{
		fmaskp += dpitch;
		dmaskp += dpitch;
		srcp += pitch;
		p += pitch;
		n += pitch;
#ifdef DEINTERLACE_MMX_BUILD
		if (chroma)
			asm_deinterlace_chroma(srcp, p, n, fmaskp, dmaskp, thresh+1, dthresh+1, w/4);
		else
			asm_deinterlace(srcp, p, n, fmaskp, dmaskp, thresh+1, dthresh+1, w/4);
#else
		for (x = 0; x < w; x += 4)
		{
			val = ((long)p[x] - (long)srcp[x]) * ((long)n[x] - (long)srcp[x]);
			if (val > thresh)
				*(int *)(fmaskp+x) = 0x000000ff;
			else
				*(int *)(fmaskp+x) = 0;
			if (val > dthresh)
				*(int *)(dmaskp+x) = 0x000000ff;
			else
				*(int *)(dmaskp+x) = 0;

			val = ((long)p[x+2] - (long)srcp[x+2]) * ((long)n[x+2] - (long)srcp[x+2]);
			if (val > thresh)
				*(int *)(fmaskp+x) |= 0x00ff0000;
			if (val > dthresh)
				*(int *)(dmaskp+x) |= 0x00ff0000;

			if (chroma)
			{
				val  = ((long)p[x+1] - (long)srcp[x+1]) * ((long)n[x+1] - (long)srcp[x+1]);
				val2 = ((long)p[x+3] - (long)srcp[x+3]) * ((long)n[x+3] - (long)srcp[x+3]);
				if (val > dthresh || val2 > dthresh)
					*(int *)(dmaskp+x) |= 0x00ff00ff;
				if (val > thresh || val2 > thresh)
					*(int *)(fmaskp+x) |= 0x00ff00ff;
			}
		}
#endif
	}
#ifdef DEINTERLACE_MMX_BUILD
	__asm emms;
#endif
	fmaskp += dpitch;
	memset(fmaskp, 0, w);
	dmaskp += dpitch;
	for (x = 0; x < w; x += 4) *(int *)(dmaskp+x) = 0x00ff00ff;

	if (full == true) return(true);

	// interlaced frame detection
	boxarraysize = ((h+8)>>3) * ((w+2)>>3);
	if ((counts = (int *) malloc(boxarraysize << 2)) == 0)
	{
		env->ThrowError("FieldDeinterlace: malloc failure");
	}
	memset(counts, 0, boxarraysize << 2);

	fmaskp = fmaskp_saved;
	p = fmaskp - dpitch;
	n = fmaskp + dpitch;
	for (y = 1; y < h - 1; y++)
	{
		boxy = (y >> 3) * (w >> 3);
		fmaskp += dpitch;
		p += dpitch;
		n += dpitch;
		for (x = 0; x < w; x += 4)
		{
			box =  boxy + (x >> 3);
			if (fmaskp[x] == 0xff && p[x] == 0xff && n[x] == 0xff)
			{
				counts[box]++;
			}

			if (fmaskp[x+2] == 0xff && p[x+2] == 0xff && n[x+2] == 0xff)
			{
				counts[box]++;
			}
		}
	}

	for (x = 0; x < boxarraysize; x++)
	{
		if (counts[x] > CT)
		{
			if (debug == true)
			{
				sprintf(buf,"FieldDeinterlace: %d: Combed frame!\n", frame);
				OutputDebugString(buf);
			}
			free(counts);
			return(true);
		}
	}
	free(counts);
	return(false);
}

PVideoFrame __stdcall FieldDeinterlace::GetFrame(int n, IScriptEnvironment* env)
{
	int x, y;
	int stat;
#ifdef BLEND_MMX_BUILD
int asm_blend(const unsigned char *dstp, const unsigned char *cprev,
								const unsigned char *cnext, unsigned char *finalp,
								unsigned char *dmaskp, int count);
#else
	unsigned int p0;
#endif
#ifdef INTERPOLATE_MMX_BUILD
int asm_interpolate(const unsigned char *dstp, const unsigned char *cprev,
								const unsigned char *cnext, unsigned char *finalp,
								unsigned char *dmaskp, int count);
#endif
	unsigned int p1, p2;
	const unsigned char *cprev, *cnext;
	char buf[255];

	src = child->GetFrame(n, env);
	srcp_saved = srcp = src->GetReadPtr();
	pitch = src->GetPitch();
	w = src->GetRowSize();
	h = src->GetHeight();

	fmask = env->NewVideoFrame(vi);
	fmaskp_saved = fmaskp = fmask->GetWritePtr();
	dpitch = fmask->GetPitch();

	dmask = env->NewVideoFrame(vi);
	dmaskp_saved = dmaskp = dmask->GetWritePtr();

	/* Check for manual overrides of the deinterlacing. */
	overrides_p = overrides;
	stat = 0;
	if (overrides_p != NULL)
	{
		while (*overrides_p < 0xffffffff)
		{
			if ((unsigned int) n >= *overrides_p && (unsigned int) n <= *(overrides_p+1) &&
				(*(overrides_p+2) == '+' || *(overrides_p+2) == '-'))
			{
				overrides_p += 2;
				stat = *overrides_p;
				break;
			}
			overrides_p += 3;
		}
		if (stat == '-')
		{
			sprintf(buf,"FieldDeinterlace: frame %d: forcing frame as not combed\n", n);
			OutputDebugString(buf);
		}
		else if (stat == '+')
		{
			sprintf(buf,"FieldDeinterlace: frame %d: forcing frame as combed\n", n);
			OutputDebugString(buf);
		}
	}

	if ((stat == '-') || (MotionMask_YUY2(n, env) == false && full == false && stat != '+'))
	{
		if (show == true)
		{
			env->MakeWritable(&src);
			sprintf(buf, "FieldDeinterlace %s", VERSION);
			DrawString(src, 0, 0, buf);
			sprintf(buf, "Copyright 2002 Donald Graft");
			DrawString(src, 0, 1, buf);
			if (stat == '-')
				sprintf(buf, "%d: Clean frame [forced]", n);
			else
				sprintf(buf, "%d: Clean frame", n);
			DrawString(src, 0, 3, buf);
		}
		return src;  // clean frame; return original
	}

	/* Frame is combed or user wants all frames, so deinterlace. */
	PVideoFrame final = env->NewVideoFrame(vi);

	srcp = srcp_saved;
	dmaskp = dmaskp_saved;
	finalp = final->GetWritePtr();
	cprev = srcp - pitch;
	cnext = srcp + pitch;
	if (blend)
	{
		/* First line. */
		for (x=0; x<w; x+=4)
		{
			p1 = *(int *)(srcp+x) & 0xfefefefe;
			p2 = *(int *)(cnext+x) & 0xfefefefe;
			*(int *)(finalp+x) = (p1>>1) + (p2>>1);
		}
		srcp += pitch;
		dmaskp += dpitch;
		finalp += dpitch;
		cprev += pitch;
		cnext += pitch;
		for (y = 1; y < h - 1; y++)
		{
#ifdef BLEND_MMX_BUILD
			asm_blend(srcp, cprev, cnext, finalp, dmaskp, w/8);
#else
			for (x=0; x<w; x+=4)
			{
				if (dmaskp[x])
				{
					p0 = *(int *)(srcp+x) & 0xfefefefe;
					p1 = *(int *)(cprev+x) & 0xfcfcfcfc;
					p2 = *(int *)(cnext+x) & 0xfcfcfcfc;
					*(int *)(finalp+x) = (p0>>1) + (p1>>2) + (p2>>2);
				}
				else
				{
					*(int *)(finalp+x)=*(int *)(srcp+x);
				}
			}
#endif
			srcp += pitch;
			dmaskp += dpitch;
			finalp += dpitch;
			cprev += pitch;
			cnext += pitch;
		}
        /* Last line. */
		for (x=0; x<w; x+=4)
		{
			p1 = *(int *)(srcp+x) & 0xfefefefe;
			p2 = *(int *)(cprev+x) & 0xfefefefe;
			*(int *)(finalp+x) = (p1>>1) + (p2>>1);
		}
#ifdef BLEND_MMX_BUILD
		__asm emms;
#endif
	}
	else
	{
		for (y = 0; y < h - 1; y++)
		{
			if (!(y&1))
			{
				memcpy(finalp,srcp,w);
			}
			else
			{
#ifdef INTERPOLATE_MMX_BUILD
			asm_interpolate(srcp, cprev, cnext, finalp, dmaskp, w/8);
#else
				for (x = 0; x < w; x += 4)
				{
					if (dmaskp[x])
					{
						p1 = *(int *)(cprev+x) & 0xfefefefe;
						p2 = *(int *)(cnext+x) & 0xfefefefe;
						*(int *)(finalp+x) = (p1>>1) + (p2>>1);
					}
					else
					{
						*(int *)(finalp+x) = *(int *)(srcp+x);
					}
				}
#endif
			}
			srcp += pitch;
			dmaskp += dpitch;
			finalp += dpitch;
			cprev += pitch;
			cnext += pitch;
		}
		/* Last line. */
		memcpy(finalp,srcp,w);
#ifdef INTERPOLATE_MMX_BUILD
		__asm emms;
#endif
	}

	if (show == true)
	{
		sprintf(buf, "FieldDeinterlace %s", VERSION);
		DrawString(final, 0, 0, buf);
		sprintf(buf, "Copyright 2002 Donald Graft");
		DrawString(final, 0, 1, buf);
		if (full == true)
		{
			sprintf(buf, "Deinterlacing all frames");
			DrawString(final, 0, 3, buf);
		}
		else
		{
			if (stat == '+')
				sprintf(buf,"%d: Combed frame! [forced]", n);
			else
				sprintf(buf,"%d: Combed frame!", n);
			DrawString(final, 0, 3, buf);
		}
	}
	
	return final;
}

AVSValue __cdecl Create_FieldDeinterlace(AVSValue args, void* user_data, IScriptEnvironment* env)
{
	char path[1024];
	char buf[80], *p;
	bool full = true;
	int threshold = 15;
	int dthreshold = 9;
	bool blend = true;
	bool chroma = false;
	char ovr[255], *ovrp;
	bool show = false;
	bool debug = false;

	try
	{
		FILE *f;

		const char* plugin_dir = env->GetVar("$PluginDir$").AsString();
		strcpy(path, plugin_dir);
		strcat(path, "\\FieldDeinterlace.def");
		if ((f = fopen(path, "r")) != NULL)
		{
			while(fgets(buf, 80, f) != 0)
			{
				if (strncmp(buf, "full=", 5) == 0)
				{
					p = buf;
					while(*p++ != '=');
					if (*p == 't') full = true;
					else full = false;
				}
				if (strncmp(buf, "threshold=", 10) == 0)
				{
					p = buf;
					while(*p++ != '=');
					threshold = atoi(p);
				}
				if (strncmp(buf, "dthreshold=", 11) == 0)
				{
					p = buf;
					while(*p++ != '=');
					dthreshold = atoi(p);
				}
				if (strncmp(buf, "blend=", 6) == 0)
				{
					p = buf;
					while(*p++ != '=');
					if (*p == 't') blend = true;
					else blend = false;
				}
				if (strncmp(buf, "chroma=", 7) == 0)
				{
					p = buf;
					while(*p++ != '=');
					if (*p == 't') chroma = true;
					else chroma = false;
				}
				ovrp = ovr;
				*ovrp = 0;
				if (strncmp(buf, "ovr=", 4) == 0)
				{
					p = buf;
					while(*p++ != '"');
					while(*p != '\n' && *p != '\r' && *p != 0 && *p != EOF && *p != '"') *ovrp++ = *p++;
					*ovrp = 0;
				}
				if (strncmp(buf, "show=", 5) == 0)
				{
					p = buf;
					while(*p++ != '=');
					if (*p == 't') show = true;
					else show = false;
				}
				if (strncmp(buf, "debug=", 6) == 0)
				{
					p = buf;
					while(*p++ != '=');
					if (*p == 't') debug = true;
					else debug = false;
				}
			}
			fclose(f);
		}
	}
	catch (...)
	{
		// plugin directory not set
		// probably using an older version avisynth
	}

	return new FieldDeinterlace(
		args[0].AsClip(),		// clip
		args[1].AsBool(full),   // full -- deinterlace all frames
		args[2].AsInt(threshold),		// threshold
		args[3].AsInt(dthreshold),		// dthreshold
		args[4].AsBool(blend),	// blend rather than interpolate
		args[5].AsBool(chroma),	// use chroma for deinterlacing
		args[6].AsString(ovr),	// overrides file
		args[7].AsBool(show),	// show debug info on frames
		args[8].AsBool(debug),  // debug mode
		env);
}
