/*
	Telecide plugin for Avisynth -- recovers original progressive
	frames from a telecined stream. The filter operates by matching
	fields and automatically adapts to phase/pattern changes. It
	operates on any arbitrary telecine pattern, e.g., 3:2 NTSC or
	1:1 PAL. An optional postprocessing phase deinterlaces frames
	that emerge from the field matching still combed.

	Copyright (C) 2002 Donald A. Graft

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	The author can be contacted at:
	Donald Graft
	neuron2@attbi.com.
*/
#include <stdio.h>
#include <windows.h>
#ifdef AVISYNTH25
#include "internal.h"
#else
#include "avisynth.h"
#endif
#include "version.h"

extern void DrawString(PVideoFrame &dst, int x, int y, const char *s); 

#define CH(n) n == 0 ? 'p' : (n == 1 ? 'c' : 'n')

struct STORE
{
	int frame;
	int choice;
};

class Telecide : public GenericVideoFilter
{
	bool reverse, swap, firstlast, agg, postprocess, blend, chroma, show, debug;
	int mm, gthresh, nt, threshold, dthreshold, y0, y1, guide;
	const char *ovr;
	int pitch, pitchtimes4, dpitch, w, h;
	unsigned int *overrides, *overrides_p;

	// Used by field matching.
	PVideoFrame fp, fc, fn, dst;
	unsigned char *fprp, *fcrp, *fcrp_saved, *fnrp;
    unsigned char *dstp, *dstp_saved;
	struct STORE store[6];
	int predicted;

	// Used by postprocessing.
	PVideoFrame fmask, dmask, final;
	const unsigned char *p, *n;
	unsigned char *fmaskp, *fmaskp_saved, *dmaskp, *dmaskp_saved, *finalp;
	bool clean;

	bool Analyse(int frame, IScriptEnvironment* env);

public:
	Telecide::Telecide(PClip _child, int _mm, bool _reverse, bool _swap, bool _firstlast, int _guide,
					   int _gthresh, bool _agg, int _nt, bool _postprocess, int _threshold,
					   int _dthreshold, bool _blend,
					   bool _chroma, int _y0, int _y1, const char * _ovr,
					   bool _show, bool _debug, IScriptEnvironment* env) :
					   GenericVideoFilter(_child), mm(_mm), reverse(_reverse), swap(_swap), firstlast(_firstlast),
					   guide(_guide), gthresh(_gthresh), agg(_agg), nt(_nt), postprocess(_postprocess),
					   threshold(_threshold), dthreshold(_dthreshold), blend(_blend), chroma(_chroma),
					   y0(_y0), y1(_y1), ovr(_ovr), show(_show), debug(_debug)
	{
		FILE *f;
		int count;
		char buf[80];
		char *d, *dsaved;
		unsigned int *p, *x;

		if (!vi.IsYUY2())
			env->ThrowError("Telecide: YUY2 data only (no RGB); use ConvertToYUY2");

		/* Load manual overrides file if there is one. */
		overrides = NULL;
		if (ovr && (f = fopen(ovr, "r")) != NULL)
		{
			/* There's an overrides file and it can be opened.
			   Get the number of lines in the file and use that to malloc
			   an in-memory data structure. The structure will have 4 ints
			   per line in the file: starting frame, ending frame, pointer to
			   string of match data, single match specification (if pointer
			   is 0). */
			count = 0;
			while(fgets(buf, 80, f) != 0)
			{
				if (buf[0] == 0 || buf[0] == '\n' || buf[0] == '\r') continue;
				count++;
			}
			fclose(f);
			if ((f = fopen(ovr, "r")) != NULL)
			{
				// Allocate in-memory store and initialize it. Leave room for end marker.
				p = overrides = (unsigned int *) malloc(4 * (count + 1) * sizeof(unsigned int));
				memset(overrides, 0xff, 4 * (count + 1) * sizeof(unsigned int));
				// Parse the file.
				while(fgets(buf, 80, f) != 0)
				{
					// Ignore blank lines.
					if (buf[0] == 0 || buf[0] == '\n' || buf[0] == '\r') continue;
					d = buf;
					// First see if a range is specified.
					while (*d != ',' && *d != 0) d++;
					if (*d == 0)
					{
						// Range not specified. Load the frame number into
						// the starting and ending frame fields.
						sscanf(buf, "%d", p);
						*(p+1) = *p;
						// Skip to pointer field.
						p += 2;
					}
					else
					{
						// Range specified. Get the starting and ending frame numbers.
						sscanf(buf, "%d,%d", p, p+1);
						// Skip to pointer field.
						p += 2;
					}
					// Determine whether a single match or a multiple match sequence is specified.
					// Skip to the first specifier and then see if is followed by another one.
					d = buf;
					while (*d != '+' && *d != '-' && *d != 'p' && *d != 'c' && *d != 'n' && *d != 0) d++;
					if (*d != 0 && *(d+1) == 'p' || *(d+1) == 'c' || *(d+1) == 'n')
					{
						// Multiple match sequence found.
						// Count the number of match specifiers.
						dsaved = d;
						count = 0;
						while (*d == 'p' || *d == 'c' || *d == 'n')
						{
							d++;
							count++;
						}
						// Get storage for the match string plus a count field.
						x = (unsigned int *) malloc(count+1);
						// Store the pointer to the match string.
						*p = (unsigned int) x;
						// Save the count first.
						*x++ = count;
						// Now save the match specifiers.
						d = dsaved;
						while (*d == 'p' || *d == 'c' || *d == 'n')
						{
							*x++ = *d++;
						}
						// Get ready for next override line.
						p += 2;
					}
					else
					{
						// Single match found. Set the pointer to 0 and store the
						// specifier.
						*p++ = 0;
						*p++ = *d;
					}
#if 0
					sprintf(buf, "%x %x %x %c\n", *(p-4), *(p-3), *(p-2), *(p-1));
					OutputDebugString(buf);
					if (*(p-2) != 0)
					{
						x = (unsigned int *) *(p-2);
						sprintf(buf, "%d: %c %c %c %c %c\n",
							    *x, *(x+1), *(x+2), *(x+3), *(x+4), *(x+5));
						OutputDebugString(buf);
					}
#endif
				}
				fclose(f);
			}
		}
		if (debug)
		{
			char b[80];
			sprintf(b, "Telecide %s by Donald Graft, Copyright 2002\n", VERSION);
			OutputDebugString(b);
		}

		vi.field_based = false;
	}

	Telecide::~Telecide(void)
	{
		unsigned int *p;

		if (overrides != NULL)
		{
			p = overrides + 2;
			while (*p != 0 && *p != 0xffffffff)
			{
				free(p);
				p += 4;
			}
			free(overrides);
		}
	}

    PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
};

#ifdef DEINTERLACE_MMX_BUILD
#pragma warning(disable:4799)
void __declspec(naked) asm_deinterlace(const unsigned char *dstp, const unsigned char *p,
								  const unsigned char *n, unsigned char *fmask,
								  unsigned char *dmask, int thresh, int dthresh, int row_size)
{
	static const __int64 Mask0 = 0xFFFF0000FFFF0000i64;
	static const __int64 Mask2 = 0x00000000000000FFi64;
	static const __int64 Mask3 = 0x0000000000FF0000i64;
	__asm
	{
		push		ebp
		push		ebx
		push		esi
		push		edi

		mov			edx,[esp+4+16]	// dstp
		mov			ecx,[esp+8+16]	// p
		mov			ebx,[esp+12+16]	// n
		mov			eax,[esp+16+16] // fmaskp
		mov			esi,[esp+20+16] // dmaskp
		movd		mm0,[esp+24+16] // threshold
		movq		mm3,mm0
		psllq		mm3,32
		por			mm3,mm0
		movd		mm4,[esp+28+16] // dthreshold
		movq		mm5,mm4
		psllq		mm5,32
		por			mm5,mm4
		mov			ebp,[esp+32+16]	// row_size
xloop:
		pxor		mm4,mm4
		movd		mm0,[ecx]		// load p
		punpcklbw	mm0,mm4			// unpack
		movd		mm2,[edx]		// load dstp
		punpcklbw	mm2,mm4			// unpack
		movd		mm1,[ebx]		// load n
		punpcklbw	mm1,mm4			// unpack
		psubw		mm0,mm2			// pprev - curr = P
		psubw		mm1,mm2			// pnext - curr = N
		movq		mm6,Mask0
		pandn		mm6,mm0			// mm6 = 0 | P2 | 0 | P0
		pmaddwd		mm6,mm1			// mm6 = P2*N2  | P0*N0
		movq		mm0,mm6

		pcmpgtd		mm6,mm3			// mm6 = P2*N2<T | P0*N0<T
		movq		mm4,mm6
		pand		mm6,Mask2
		movq		mm7,mm6
		psrlq		mm4,32
		pand		mm4,Mask3
		por			mm7,mm4
		movd		[eax],mm7

		pcmpgtd		mm0,mm5			// mm0 = P2*N2<DT | P0*N0<DT
		movq		mm4,mm0
		pand		mm0,Mask2
		movq		mm7,mm0
		psrlq		mm4,32
		pand		mm4,Mask3
		por			mm7,mm4
		movd		[esi],mm7

		add			edx,4
		add			ecx,4
		add			ebx,4
		add			eax,4
		add			esi,4
		dec			ebp
		jnz			xloop

		pop			edi
		pop			esi
		pop			ebx
		pop			ebp
		ret
	};	
}

void __declspec(naked) asm_deinterlace_chroma(const unsigned char *dstp, const unsigned char *p,
								  const unsigned char *n, unsigned char *fmask,
								  unsigned char *dmask, int thresh, int dthresh, int row_size)
{
	static const __int64 Mask0 = 0xFFFF0000FFFF0000i64;
	static const __int64 Mask2 = 0x00000000000000FFi64;
	static const __int64 Mask3 = 0x0000000000FF0000i64;
	static const __int64 Mask4 = 0x0000000000FF00FFi64;
	__asm
	{
		push		ebp
		push		ebx
		push		esi
		push		edi

		mov			edx,[esp+4+16]	// dstp
		mov			ecx,[esp+8+16]	// p
		mov			ebx,[esp+12+16]	// n
		mov			eax,[esp+16+16] // fmaskp
		mov			esi,[esp+20+16] // dmaskp
		movd		mm0,[esp+24+16] // threshold
		movq		mm3,mm0
		psllq		mm3,32
		por			mm3,mm0
		movd		mm4,[esp+28+16] // dthreshold
		movq		mm5,mm4
		psllq		mm5,32
		por			mm5,mm4
		mov			ebp,[esp+32+16]	// row_size
xloop:
		pxor		mm4,mm4
		movd		mm0,[ecx]		// load p
		punpcklbw	mm0,mm4			// unpack
		movd		mm2,[edx]		// load dstp
		punpcklbw	mm2,mm4			// unpack
		movd		mm1,[ebx]		// load n
		punpcklbw	mm1,mm4			// unpack
		psubw		mm0,mm2			// pprev - curr = P
		psubw		mm1,mm2			// pnext - curr = N
		movq		mm6,Mask0
		pandn		mm6,mm0			// mm6 = 0 | P2 | 0 | P0
		pmaddwd		mm6,mm1			// mm6 = P2*N2  | P0*N0
		movq		mm0,mm6

		pcmpgtd		mm6,mm3			// mm6 = P2*N2<T | P0*N0<T
		movq		mm4,mm6			// save result
		pand		mm6,Mask2		// make 0x000000ff if P0*N0<T
		movq		mm7,mm6			// move it into result
		psrlq		mm4,32			// get P2*N2<T
		pand		mm4,Mask3		// make 0x00ff0000 if P2*N2<T
		por			mm7,mm4			// move it into the result
		movd		[eax],mm7		// save intermediate result

		pcmpgtd		mm0,mm5			// mm0 = P2*N2<DT | P0*N0<DT
		movq		mm4,mm0			// save result
		pand		mm0,Mask2		// make 0x000000ff if P0*N0<DT
		movq		mm7,mm0			// move it into result
		psrlq		mm4,32			// get P2*N2<DT
		pand		mm4,Mask3		// make 0x00ff0000 if P2*N2<DT
		por			mm7,mm4			// move it into the result
		movd		[esi],mm7		// save intermediate result

		pxor		mm4,mm4			// do chroma bytes now
		movd		mm0,[ecx]		// load p
		punpcklbw	mm0,mm4			// unpack
		movd		mm1,[ebx]		// load n
		punpcklbw	mm1,mm4			// unpack
		psubw		mm0,mm2			// pprev - curr = P
		psubw		mm1,mm2			// pnext - curr = N
		movq		mm6,Mask0
		pand		mm6,mm0			// mm6 = P3 | 0 | P1| 0
		pmaddwd		mm6,mm1			// mm6 = P3*N3  | P1*N1
		movq		mm0,mm6

		pcmpgtd		mm6,mm3			// mm6 = P3*N3<T | P1*N1<T
		movq		mm4,mm6			// save result
		pand		mm6,Mask4		// make 0x00ff00ff if P1*N1<T
		movq		mm7,mm6			// move it into result
		psrlq		mm4,32			// get P3*N3<T
		pand		mm4,Mask4		// make 0x00ff00ff if P3*N3<T
		por			mm7,mm4			// move it into the result
		movd		mm4,[eax]		// Pick up saved intermediate result
		por			mm4,mm7
		movd		[eax],mm4		// Save final result for fmask

		pcmpgtd		mm0,mm5			// mm0 = P3*N3<DT | P1*N1<DT
		movq		mm4,mm0			// save result
		pand		mm0,Mask4		// make 0x00ff00ff if P1*N1<DT
		por			mm7,mm0			// move it into result
		psrlq		mm4,32			// get P3*N3<DT
		pand		mm4,Mask4		// make 0x00ff00ff if P3*N3<DT
		por			mm7,mm4			// move it into the result
		movd		mm4,[esi]		// Pick up saved intermediate result
		por			mm4,mm7
		movd		[esi],mm4		// Save final result for dmask

		add			edx,4
		add			ecx,4
		add			ebx,4
		add			eax,4
		add			esi,4
		dec			ebp
		jnz			xloop

		pop			edi
		pop			esi
		pop			ebx
		pop			ebp
		ret
	};	
}
#pragma warning(default:4799)
#endif

#ifdef MATCH_MMX_BUILD
#pragma warning(disable:4799)
int __declspec(naked) asm_match(const unsigned char *curr, const unsigned char *pprev,
								  const unsigned char *pnext, int nt, int row_size)
{
	static const __int64 Mask = 0xFFFF0000FFFF0000i64;
	static const __int64 Ones = 0x0000000100000001i64;
	__asm
	{
		push ebp
		push ebx

		mov			edx,[esp+4+8]	// curr
		mov			ecx,[esp+8+8]	// pprev
		mov			ebx,[esp+12+8]	// pnext
		mov			ebp,[esp+20+8]	// rowsize
		movd		mm0,[esp+16+8] // nt
		movq		mm7,mm0
		psllq		mm7,32
		por			mm7,mm0
		movq		mm6,Mask
		pxor		mm3,mm3			// running sum
xloop:
		movd		mm0,[ecx]		// load pprev
		pxor		mm4,mm4

		movd		mm2,[edx]		// load curr
		punpcklbw	mm0,mm4			// unpack

		movd		mm1,[ebx]		// load pnext
		punpcklbw	mm2,mm4			// unpack

		punpcklbw	mm1,mm4			// unpack
		psubw		mm0,mm2			// pprev - curr

		psubw		mm1,mm2			// pnext - curr
		movq		mm4,mm6

		pandn		mm6,mm0
		pand		mm0,mm4
		pmaddwd		mm0,mm1			// mm1 = P3*N3  | P1*N1
		pmaddwd		mm6,mm1			// mm6 = P2*N2  | P0*N0
		movq		mm5,mm7
		movq		mm2,mm7
		pcmpgtd		mm2,mm0
		pcmpgtd		mm5,mm6

		movq		mm6,Ones
		pandn		mm5,mm6
		pandn		mm2,mm6
		paddd		mm3,mm5	
		paddd		mm3,mm2
		movq		mm6,mm4

		add			edx,16
		add			ecx,16
		add			ebx,16

		dec			ebp
		jnz			xloop

		movq		mm0,mm3
		psrlq		mm3,32
		paddd		mm0,mm3
		movd		eax,mm0
		pop			ebx
		pop			ebp
		ret
	};
}
#pragma warning(default:4799)
#endif

#ifdef BLEND_MMX_BUILD
#pragma warning(disable:4799)
int __declspec(naked) asm_blend(const unsigned char *dstp, const unsigned char *cprev,
								const unsigned char *cnext, unsigned char *finalp,
								unsigned char *dmaskp, int count)
{
	static const __int64 Mask1 = 0xFEFEFEFEFEFEFEFEi64;
	static const __int64 Mask2 = 0xFCFCFCFCFCFCFCFCi64;
	__asm
	{
		push ebp
		push ebx
		push edi
		push esi

		mov			edx,[esp+4+16]	// dstp
		mov			ecx,[esp+8+16]	// cprev
		mov			ebx,[esp+12+16]	// cnext
		mov			eax,[esp+16+16] // finalp
		mov			edi,[esp+20+16]	// dmaskp
		mov			ebp,[esp+24+16]	// count
		movq		mm7,Mask1
		movq		mm6,Mask2
xloop:
		mov			esi,[edi]
		or			esi,esi
		jnz			blend
		mov			esi,[edi+4]
		or			esi,esi
		jnz			blend
		movq		mm0,[edx]
		movq		[eax],mm0
		jmp			skip
blend:
		movq		mm0,[edx]		// load dstp
		pand		mm0,mm7

		movq		mm1,[ecx]		// load cprev
		pand		mm1,mm6

		movq		mm2,[ebx]		// load cnext
		pand		mm2,mm6

		psrlq		mm0,1
		psrlq		mm1,2
		psrlq       mm2,2
		paddusb		mm0,mm1
		paddusb		mm0,mm2
		movq		[eax],mm0
skip:
		add			edx,8
		add			ecx,8
		add			ebx,8
		add			eax,8
		add			edi,8
		dec			ebp
		jnz			xloop

		pop			esi
		pop			edi
		pop			ebx
		pop			ebp
		ret
	};
}
#pragma warning(default:4799)
#endif

#ifdef INTERPOLATE_MMX_BUILD
#pragma warning(disable:4799)
int __declspec(naked) asm_interpolate(const unsigned char *dstp, const unsigned char *cprev,
								const unsigned char *cnext, unsigned char *finalp,
								unsigned char *dmaskp, int count)
{
	static const __int64 Mask = 0xFEFEFEFEFEFEFEFEi64;
	__asm
	{
		push ebp
		push ebx
		push edi
		push esi

		mov			edx,[esp+4+16]	// dstp
		mov			ecx,[esp+8+16]	// cprev
		mov			ebx,[esp+12+16]	// cnext
		mov			eax,[esp+16+16] // finalp
		mov			edi,[esp+20+16]	// dmaskp
		mov			ebp,[esp+24+16]	// count
		movq		mm7,Mask
xloop:
		mov			esi,[edi]
		or			esi,esi
		jnz			blend
		mov			esi,[edi+4]
		or			esi,esi
		jnz			blend
		movq		mm0,[edx]
		movq		[eax],mm0
		jmp			skip
blend:
		movq		mm0,[ecx]		// load cprev
		pand		mm0,mm7

		movq		mm1,[ebx]		// load cnext
		pand		mm1,mm7

		psrlq		mm0,1
		psrlq       mm1,1
		paddusb		mm0,mm1
		movq		[eax],mm0
skip:
		add			edx,8
		add			ecx,8
		add			ebx,8
		add			eax,8
		add			edi,8
		dec			ebp
		jnz			xloop

		pop			esi
		pop			edi
		pop			ebx
		pop			ebp
		ret
	};
}
#pragma warning(default:4799)
#endif

bool Telecide::Analyse(int frame, IScriptEnvironment* env)
{
	int x, y, thresh, dthresh;
	int *counts, box, boxy;
#ifndef DEINTERLACE_MMX_BUILD
	int val, val2;
#endif
	const CT = 15;
	int boxarraysize;

	thresh = threshold*threshold;
	dthresh = dthreshold*dthreshold;

	p = dstp - dpitch;
	n = dstp + dpitch;
	memset(fmaskp, 0, w);
	for (x = 0; x < w; x += 4) *(int *)(dmaskp+x) = 0x00ff00ff;

	for (y = 1; y < h - 1; y++)
	{
		fmaskp += dpitch;
		dmaskp += dpitch;
		dstp += dpitch;
		p += dpitch;
		n += dpitch;
#ifdef DEINTERLACE_MMX_BUILD
		if (chroma)
			asm_deinterlace_chroma(dstp, p, n, fmaskp, dmaskp, thresh+1, dthresh+1, w/4);
		else
			asm_deinterlace(dstp, p, n, fmaskp, dmaskp, thresh+1, dthresh+1, w/4);
#else
		for (x = 0; x < w; x += 4)
		{
			val = ((long)p[x] - (long)dstp[x]) * ((long)n[x] - (long)dstp[x]);
			if (val > thresh)
				*(int *)(fmaskp+x) = 0x000000ff;
			else
				*(int *)(fmaskp+x) = 0;
			if (val > dthresh)
				*(int *)(dmaskp+x) = 0x000000ff;
			else
				*(int *)(dmaskp+x) = 0;

			val = ((long)p[x+2] - (long)dstp[x+2]) * ((long)n[x+2] - (long)dstp[x+2]);
			if (val > thresh)
				*(int *)(fmaskp+x) |= 0x00ff0000;
			if (val > dthresh)
				*(int *)(dmaskp+x) |= 0x00ff0000;

			if (chroma)
			{
				val  = ((long)p[x+1] - (long)dstp[x+1]) * ((long)n[x+1] - (long)dstp[x+1]);
				val2 = ((long)p[x+3] - (long)dstp[x+3]) * ((long)n[x+3] - (long)dstp[x+3]);
				if (val > dthresh || val2 > dthresh)
					*(int *)(dmaskp+x) |= 0x00ff00ff;
				if (val > thresh || val2 > thresh)
					*(int *)(fmaskp+x) |= 0x00ff00ff;
			}
		}
#endif
	}
#ifdef DEINTERLACE_MMX_BUILD
	__asm emms;
#endif
	fmaskp += dpitch;
	memset(fmaskp, 0, w);
	dmaskp += dpitch;
	for (x = 0; x < w; x += 4) *(int *)(dmaskp+x) = 0x00ff00ff;

	// interlaced frame detection
	boxarraysize = ((h+8)>>3) * ((w+2)>>3);
	if ((counts = (int *) malloc(boxarraysize << 2)) == 0)
	{
		env->ThrowError("Telecide: malloc failure");
	}
	memset(counts, 0, boxarraysize << 2);

	fmaskp = fmaskp_saved;
	p = fmaskp - dpitch;
	n = fmaskp + dpitch;
	for (y = 1; y < h - 1; y++)
	{
		boxy = (y >> 3) * (w >> 3);
		fmaskp += dpitch;
		p += dpitch;
		n += dpitch;
		for (x = 0; x < w; x += 4)
		{
			box =  boxy + (x >> 3);
			if (fmaskp[x] == 0xff && p[x] == 0xff && n[x] == 0xff)
			{
				counts[box]++;
			}

			if (fmaskp[x+2] == 0xff && p[x+2] == 0xff && n[x+2] == 0xff)
			{
				counts[box]++;
			}
		}
	}

	for (x = 0; x < boxarraysize; x++)
	{
		if (counts[x] > CT)
		{
			free(counts);
			return(true);
		}
	}
	free(counts);
	return(false);
}

void DrawShow(PVideoFrame &dst, int chosen, int p, int c, int n, int frame, bool found,
			  bool pattern_override, int guide, struct STORE *store)
{
	char l, buf[80];

	sprintf(buf, "Telecide %s", VERSION);
	DrawString(dst, 0, 0, buf);
	sprintf(buf, "Copyright 2002 Donald Graft");
	DrawString(dst, 0, 1, buf);
	if (chosen == 0) l = 'p';
	else if (chosen == 1) l = 'c';
	else l = 'n';
	sprintf(buf,"frame %d:", frame);
	DrawString(dst, 0, 3, buf);
	if (found == true)
		sprintf(buf,"frame %d: manual field match [using %c]\n", frame, l);
	else
		sprintf(buf,"p=%u  c=%u  n=%u [using %c]", p, c, n, l);
	DrawString(dst, 0, 4, buf);
	if (guide == 1)
	{
		sprintf(buf,"[%c %c %c %c] %c %s",
				CH(store[0].choice), CH(store[1].choice), CH(store[2].choice),
				CH(store[3].choice), CH(store[4].choice),
				pattern_override == true ? "...pattern override" : "");
		DrawString(dst, 0, 5, buf);
	}
	else if (guide == 2)
	{
		sprintf(buf,"[%c %c] %c %s",
				CH(store[0].choice), CH(store[1].choice), CH(store[2].choice),
				pattern_override == true ? "...pattern override" : "");
		DrawString(dst, 0, 6, buf);
	}
	else if (guide == 3)
	{
		sprintf(buf,"[%c %c %c %c %c] %c %s",
				CH(store[0].choice), CH(store[1].choice), CH(store[2].choice),
				CH(store[3].choice), CH(store[4].choice), CH(store[5].choice),
				pattern_override == true ? "...pattern override" : "");
		DrawString(dst, 0, 5, buf);
	}
}

PVideoFrame __stdcall Telecide::GetFrame(int frame, IScriptEnvironment* env)
{
	int p, c, n, lowest;
#ifndef MATCH_MMX_BUILD
	int C;
	int comb;
#endif
    unsigned char *curr, *pprev, *pnext, *cprev, *cnext, *nprev, *nnext;
#ifndef BLEND_MMX_BUILD
	unsigned int p0;
#endif
	unsigned int p1, p2;
    int hminus2, hover2, hover4minus2, hplus1over2;
	int x, y;
	int chosen;
	char buf[80];
	bool found;
	int force;
	bool pattern_override;

	/* Do field matching. */
	fp = child->GetFrame(frame == 0 ? 0 : frame - 1, env);
	fc = child->GetFrame(frame, env);
	fn = child->GetFrame(frame >= vi.num_frames - 1 ? vi.num_frames - 1: frame + 1, env);
	dst = env->NewVideoFrame(vi);

	pitch = fc->GetPitch();
	pitchtimes4 = pitch << 2;
	dpitch = dst->GetPitch();
	w = fc->GetRowSize();
	if ((w/2) % 4)
		env->ThrowError("Telecide: width must be a multiple of 4; use Crop");
	h = fc->GetHeight();
	if (y0 < 0 || y1 < 0 || y0 > y1)
		env->ThrowError("Telecide: bad y0/y1 values");
	hminus2 = h - 2;
	hover2 = h/2;
	hover4minus2 = h/4 - 2;
	hplus1over2 = (h+1)/2;

	fprp = (unsigned char *) fp->GetReadPtr();
	fcrp = (unsigned char *) fc->GetReadPtr();
	fnrp = (unsigned char *) fn->GetReadPtr();

	/* Check for manual overrides of the field matching. */
	overrides_p = overrides;
	found = false;
	if (overrides_p != NULL)
	{
		while (*overrides_p < 0xffffffff)
		{
			// If the frame is in range...
			if (((unsigned int) frame >= *overrides_p) && ((unsigned int) frame <= *(overrides_p+1)))
			{
				// and it's a single specifier. 
				if (*(overrides_p+2) == 0)
				{
					// and it's a matching specifier.
					if (*(overrides_p+3) == 'p' || *(overrides_p+3) == 'c' || *(overrides_p+3) == 'n')
					{
						// Get the match specifier and stop parsing.
						found = true;
						overrides_p += 3;
						switch(*overrides_p)
						{
						case 'p': chosen = 0; break;
						case 'c': chosen = 1; break;
						case 'n': chosen = 2; break;
						}
						break;
					}
				}
				else
				{
					// It's a multiple match specifier.
					found = true;
					// Get the pointer to the specifier string.
					unsigned int *x = (unsigned int *) *(overrides_p+2);
					// Get the index into the specification string.
					// Remember, the count is first followed by the specifiers.
					int ndx = ((frame - *overrides_p) % *x);
					// Point to the specifier string.
					x++;
					// Load the specifier for this frame and stop parsing.
					switch(x[ndx])
					{
					case 'p': chosen = 0; break;
					case 'c': chosen = 1; break;
					case 'n': chosen = 2; break;
					}
					break;
				}
			}
			// Next override line.
			overrides_p += 4;
		}
	}

	if (found == true)
	{
		/* Manual override found. */
		if (guide == 1)
		{
			store[0] = store[1];
			store[1] = store[2];
			store[2] = store[3];
			store[3] = store[4];
			store[4].frame = frame;
			store[4].choice = chosen;
		}
		else if (guide == 2)
		{
			store[0] = store[1];
			store[1] = store[2];
			store[2].frame = frame;
			store[2].choice = chosen;
		}
		else if (guide == 3)
		{
			store[0] = store[1];
			store[1] = store[2];
			store[2] = store[3];
			store[3] = store[4];
			store[4] = store[5];
			store[5].frame = frame;
			store[5].choice = chosen;
		}
	}
	else
	{
		/* No manual override. */
		if (reverse == true)
		{
			curr =  (unsigned char *) (fcrp + 2 * pitch);
			pprev = (unsigned char *) (fprp + pitch);
			pnext = (unsigned char *) (fprp + 3 * pitch);
			cprev = (unsigned char *) (fcrp + pitch);
			cnext = (unsigned char *) (fcrp + 3 * pitch);
			nprev = (unsigned char *) (fnrp + pitch);
			nnext = (unsigned char *) (fnrp + 3 * pitch);
		}
		else
		{
			curr =  (unsigned char *) (fcrp + pitch);
			pprev = (unsigned char *) (fprp);
			pnext = (unsigned char *) (fprp + 2 * pitch);
			cprev = (unsigned char *) (fcrp);
			cnext = (unsigned char *) (fcrp + 2 * pitch);
			nprev = (unsigned char *) (fnrp);
			nnext = (unsigned char *) (fnrp + 2 * pitch);
		}
		p = c = n = 0;
		/* Try to match the top field of the current frame to the
		   bottom fields of the previous, current, and next frames.
		   Output the assembled frame that matches up best. For
		   matching, subsample the frames in the x dimension
		   for speed. */
		for (y = 0; y < hminus2; y+=4)
		{
			/* Exclusion band. Good for ignoring subtitles. */
			if (y0 == y1 || y < y0 || y > y1)
			{
#ifdef MATCH_MMX_BUILD
				if (mm == 0 || mm == 1) p += asm_match(curr, pprev, pnext, nt+1, w / 16);
				c += asm_match(curr, cprev, cnext, nt+1, w / 16);
				if (mm == 0 || mm == 2) n += asm_match(curr, nprev, nnext, nt+1, w / 16);
#else
				for (x = 0; x < w;)
				{
					C = curr[x];
					/* This combing metric is based on an original idea of Gunnar Thalin. */
					if (mm == 0 || mm == 1)
					{
						comb = ((long)pprev[x] - C) * ((long)pnext[x] - C);
						if (comb > nt) p++;
					}

					comb = ((long)cprev[x] - C) * ((long)cnext[x] - C);
					if (comb > nt) c++;

					if (mm == 0 || mm == 2)
					{
						comb = ((long)nprev[x] - C) * ((long)nnext[x] - C);
						if (comb > nt) n++;
					}

					x++;
					if (!(x&3)) x += 12;
				}
#endif
			}
			curr  += pitchtimes4;
			pprev += pitchtimes4;
			pnext += pitchtimes4;
			cprev += pitchtimes4;
			cnext += pitchtimes4;
			nprev += pitchtimes4;
			nnext += pitchtimes4;
		}
#ifdef MATCH_MMX_BUILD
		__asm emms;
#endif

		/* The idea of guiding blind field matching with the prevailing
		   pattern was pioneered by Alan Liu ("daxab"). */
		if (guide == 1)
		{
			/* NTSC 1 guidance. 24fps -> 30fps telecine. */
			predicted = 0xff;
			if (store[4].frame == frame - 1 &&
				store[3].frame == frame - 2 &&
				store[2].frame == frame - 3 &&
				store[1].frame == frame - 4 &&
				store[0].frame == frame - 5)
			{
				switch ((store[0].choice << 16) +
						(store[1].choice << 12) +
						(store[2].choice <<  8) +
						(store[3].choice <<  4) +
						(store[4].choice))
				{
				case 0x11100:
				case 0x11001:
				case 0x10011:
					predicted = 1;
					break;
				case 0x10001: 
				case 0x01100: 
				case 0x11000:
					if (agg == true) predicted = 1;
					break;
				case 0x00111:
				case 0x01110:
					predicted = 0;
					break;
				case 0x00110: 
				case 0x00011: 
					if (agg == true) predicted = 0;
					break;
				case 0x11122:
				case 0x11221:
				case 0x12211:
					predicted = 1;
					break;
				case 0x12221: 
				case 0x21122: 
				case 0x11222: 
					if (agg == true) predicted = 1;
					break;
				case 0x22111:
				case 0x21112:
					predicted = 2;
					break;
				case 0x22112: 
				case 0x22211: 
					if (agg == true) predicted = 2;
					break;
				}
			}

			store[0] = store[1];
			store[1] = store[2];
			store[2] = store[3];
			store[3] = store[4];
			store[4].frame = frame;
		}
		else if (guide == 2)
		{
			/* PAL guidance. */
			predicted = 0xff;
			if (store[2].frame == frame - 1 &&
				store[1].frame == frame - 2 &&
				store[0].frame == frame - 3)
			{
				switch ((store[0].choice << 8) +
						(store[1].choice << 4) +
						(store[2].choice))
				{
				case 0x000:
					predicted = 0;
					break;
				case 0x111:
					predicted = 1;
					break;
				case 0x222:
					predicted = 2;
					break;
				}
			}

			store[0] = store[1];
			store[1] = store[2];
			store[2].frame = frame;
		}
		else if (guide == 3)
		{
			/* NTSC 2 guidance. 25fps -> 30fps telecine. */
			predicted = 0xff;
			if (store[5].frame == frame - 1 &&
				store[4].frame == frame - 2 &&
				store[3].frame == frame - 3 &&
				store[2].frame == frame - 4 &&
				store[1].frame == frame - 5 &&
				store[0].frame == frame - 6)
			{
				switch ((store[0].choice << 20) +
						(store[1].choice << 16) +
						(store[2].choice << 12) +
						(store[3].choice <<  8) +
						(store[4].choice <<  4) +
						(store[5].choice))
				{
				case 0x111100:
				case 0x111001:
				case 0x110011:
				case 0x100111:
					predicted = 1;
					break;
				case 0x001111:
				case 0x011110:
					predicted = 0;
					break;
				case 0x111122:
				case 0x111221:
				case 0x112211:
				case 0x122111:
					predicted = 1;
					break;
				case 0x221111:
				case 0x211112:
					predicted = 2;
					break;
				}
			}

			store[0] = store[1];
			store[1] = store[2];
			store[2] = store[3];
			store[3] = store[4];
			store[4] = store[5];
			store[5].frame = frame;
		}

		lowest = c;
		chosen = 1;
		if (guide == 1) store[4].choice = 1;
		else if (guide == 2) store[2].choice = 1;
		else if (guide == 3) store[5].choice = 1;
		if ((mm == 0 || mm == 1) && p < lowest)
		{
			lowest = p;
			chosen = 0;
			if (guide == 1) store[4].choice = 0;
			else if (guide == 2) store[2].choice = 0;
			else if (guide == 3) store[5].choice = 0;
		}
		if ((mm == 0 || mm == 2) &&n < lowest)
		{
			lowest = n;
			chosen = 2;
			if (guide == 1) store[4].choice = 2;
			else if (guide == 2) store[2].choice = 2;
			else if (guide == 3) store[5].choice = 2;
		}
		if (guide == 1 || guide == 2 || guide == 3)
		{
			if (predicted != 0xff && predicted != chosen)
			{
				int val;
				switch (predicted)
				{
				case 0: val = p; break;
				case 1: val = c; break;
				case 2: val = n; break;
				}
				/* If the predicted match metric is within defined percentage of the
				   best calculated one, then override the calculated match with the
				   predicted match. */
				pattern_override = false;
				if (abs(val - lowest) < (val * gthresh) / 100)
				{
					if (debug)
					{
						sprintf(buf,"Telecide: override %d: val=%d lowest=%d\n", frame, val, lowest);
						OutputDebugString(buf);
					}
					pattern_override = true;
					chosen = predicted;
					if (guide == 1) store[4].choice = predicted;
					else if (guide == 2) store[2].choice = predicted;
					else if (guide == 3) store[5].choice = predicted;
				}
			}
		}
	}
	if (debug)
	{
		char l;
		if (chosen == 0) l = 'p';
		else if (chosen == 1) l = 'c';
		else l = 'n';
		if (found == true)
			sprintf(buf,"Telecide: frame %d: manual field match [using %c]\n", frame, l);
		else
			sprintf(buf,"Telecide: frame %d: p=%u  c=%u  n=%u [using %c]\n", frame, p, c, n, l);
		OutputDebugString(buf);
		if (guide == 1)
		{
			sprintf(buf,"Telecide: [%c %c %c %c] %c\n",
					CH(store[0].choice), CH(store[1].choice), CH(store[2].choice),
					CH(store[3].choice), CH(store[4].choice));
			OutputDebugString(buf);
		}
		else if (guide == 2)
		{
			sprintf(buf,"Telecide: [%c %c] %c\n",
					CH(store[0].choice), CH(store[1].choice), CH(store[2].choice));
			OutputDebugString(buf);
		}
		if (guide == 3)
		{
			sprintf(buf,"Telecide: [%c %c %c %c %c] %c\n",
					CH(store[0].choice), CH(store[1].choice), CH(store[2].choice),
					CH(store[3].choice), CH(store[4].choice), CH(store[5].choice));
			OutputDebugString(buf);
		}
	}

	/* Set up the pointers in preparation to output final frame. */
	dstp_saved = dstp = dst->GetWritePtr();
	if (reverse == true) dstp += dpitch;
	if (swap == true) dstp = (dstp == dstp_saved ? dstp_saved + dpitch : dstp_saved);
	if (chosen == 0) curr = fprp;
	else if (chosen == 1) curr = fcrp;
	else curr = fnrp;
	if (reverse == true) curr += pitch;

	/* First output the bottom field selected from the set of three stored
	   frames. */
	for (y = 0; y < hplus1over2; y++)
	{
		memcpy(dstp, curr, w);
		curr += pitch << 1;
		dstp += dpitch << 1;
	}

	/* Now output the top field of the current frame. */
	if (reverse == true)
	{
		curr = (unsigned char *) (fcrp);
		dstp = dstp_saved;
	}
	else
	{
		curr = (unsigned char *) (fcrp + pitch);
		dstp = dstp_saved + dpitch;
	}
	if (swap == true) dstp = (dstp == dstp_saved ? dstp_saved + dpitch : dstp_saved);
	for (y = 0; y < hover2; y++)
	{
		memcpy(dstp, curr, w);
		curr += pitch << 1;
		dstp += dpitch << 1;
	}

	/* Check for manual overrides of the deinterlacing. */
	overrides_p = overrides;
	force = 0;
	if (overrides_p != NULL)
	{
		while (*overrides_p < 0xffffffff)
		{
			// Is the frame in range...
			if (((unsigned int) frame >= *overrides_p) && ((unsigned int) frame <= *(overrides_p+1)) &&
				// and is it a single specifier...
				(*(overrides_p+2) == 0) &&
				// and is it a deinterlacing specifier?
				(*(overrides_p+3) == '+' || *(overrides_p+3) == '-'))
			{
				// Yes, load the specifier and stop parsing.
				overrides_p += 3;
				force = *overrides_p;
				break;
			}
			// Next specification record.
			overrides_p += 4;
		}
	}

	/* It's possible to get a bad match on the first and/or last frames. Force them to be
	   postprocessed if enabled. */
	if (firstlast == true && (frame == 0 || frame == vi.num_frames - 1))
	{
		clean = false;
		force = '+';
	}

	// If not postprocessing and not forced, just finish up.
	if ((postprocess == false) && (force != '+') && (force != '-'))
	{
		if (show == true)
		{
			DrawShow(dst, chosen, p, c, n, frame, found, pattern_override, guide, store);
		}
		return dst;
	}

	clean = true;
	if (force != '-')
	{
		/* If the frame is not forced as clean by user override, then
		   process it for deinterlacing. */

		/* Setup the motion maps: fmask for determining whether the frame is
		   combed (uses threshold), and dmask for the actual deinterlacing
		   (uses dthreshold). */
		dstp = dstp_saved;
		fmask = env->NewVideoFrame(vi);
		fmaskp_saved = fmaskp = fmask->GetWritePtr();

		dmask = env->NewVideoFrame(vi);
		dmaskp_saved = dmaskp = dmask->GetWritePtr();

		final = env->NewVideoFrame(vi);
		finalp = final->GetWritePtr();

		if (Analyse(frame, env) == true || force == '+')
		{
			/* User wants all frames deinterlaced, or has forced deinterlacing
			   for this frame, or the frame is combed, so deinterlace. */
			clean = false;

			dstp = dstp_saved;
			dmaskp = dmaskp_saved;
			cprev = dstp - dpitch;
			cnext = dstp + dpitch;
			if (blend == true)
			{
				/* First line. */
				for (x=0; x<w; x+=4)
				{
					p1 = *(int *)(dstp+x) & 0xfefefefe;
					p2 = *(int *)(cnext+x) & 0xfefefefe;
					*(int *)(finalp+x) = (p1>>1) + (p2>>1);
				}
				dstp += dpitch;
				dmaskp += dpitch;
				finalp += dpitch;
				cprev += dpitch;
				cnext += dpitch;
				for (y = 1; y < h - 1; y++)
				{
#ifdef BLEND_MMX_BUILD
					asm_blend(dstp, cprev, cnext, finalp, dmaskp, w/8);
#else
					for (x=0; x<w; x+=4)
					{
						if (dmaskp[x])
						{
							p0 = *(int *)(dstp+x) & 0xfefefefe;
							p1 = *(int *)(cprev+x) & 0xfcfcfcfc;
							p2 = *(int *)(cnext+x) & 0xfcfcfcfc;
							*(int *)(finalp+x) = (p0>>1) + (p1>>2) + (p2>>2);
						}
						else
						{
							*(int *)(finalp+x)=*(int *)(dstp+x);
						}
					}
#endif
					dstp += dpitch;
					dmaskp += dpitch;
					finalp += dpitch;
					cprev += dpitch;
					cnext += dpitch;
				}
				/* Last line. */
				for (x=0; x<w; x+=4)
				{
					p1 = *(int *)(dstp+x) & 0xfefefefe;
					p2 = *(int *)(cprev+x) & 0xfefefefe;
					*(int *)(finalp+x) = (p1>>1) + (p2>>1);
				}
#ifdef BLEND_MMX_BUILD
				__asm emms;
#endif
			}
			else
			{
				for (y = 0; y < h - 1; y++)
				{
					if (!(y&1))
					{
						memcpy(finalp,dstp,w);
					}
					else
					{
#ifdef INTERPOLATE_MMX_BUILD
						asm_interpolate(dstp, cprev, cnext, finalp, dmaskp, w/8);
#else
						for (x = 0; x < w; x += 4)
						{
							if (dmaskp[x])
							{
								p1 = *(int *)(cprev+x) & 0xfefefefe;
								p2 = *(int *)(cnext+x) & 0xfefefefe;
								*(int *)(finalp+x) = (p1>>1) + (p2>>1);
							}
							else
							{
								*(int *)(finalp+x) = *(int *)(dstp+x);
							}
						}
#endif
					}
					dstp += dpitch;
					dmaskp += dpitch;
					finalp += dpitch;
					cprev += dpitch;
					cnext += dpitch;
				}
				/* Last line. */
				memcpy(finalp,dstp,w);
#ifdef INTERPOLATE_MMX_BUILD
				__asm emms;
#endif
			}
		}
	}

	// Show debugging info if enabled.
	if (clean == true)
	{
		if (show == true)
		{
			DrawShow(dst, chosen, p, c, n, frame, found, pattern_override, guide, store);
			if (force == '-')
				sprintf(buf, "%d: Clean frame [forced]", frame);
			else
				sprintf(buf, "%d: Clean frame", frame);
			DrawString(dst, 0, 7, buf);
		}
		if (debug == true)
		{
			if (force == '-')
			{
				sprintf(buf,"Telecide: frame %d: Clean frame [forced]\n", frame);
				OutputDebugString(buf);
			}
			else
			{
				sprintf(buf,"Telecide: frame %d: Clean frame\n", frame);
				OutputDebugString(buf);
			}
		}
	}
	else
	{
		if (show == true)
		{
			DrawShow(final, chosen, p, c, n, frame, found, pattern_override, guide, store);
			if (force == '+')
				sprintf(buf,"%d: Combed frame! [forced]", frame);
			else
				sprintf(buf,"%d: Combed frame!", frame);
			DrawString(final, 0, 7, buf);
		}
		if (debug == true)
		{
			if (force == '+')
			{
				sprintf(buf,"Telecide: frame %d: Combed frame! [forced]\n", frame);
				OutputDebugString(buf);
			}
			else
			{
				sprintf(buf,"Telecide: frame %d: Combed frame!\n", frame);
				OutputDebugString(buf);
			}
		}
	}

	if (clean == false)	return final;
	else return dst;
}

AVSValue __cdecl Create_Telecide(AVSValue args, void* user_data, IScriptEnvironment* env)
{
	char path[1024];
	char buf[80], *p;
	int mm = 0;
	bool reverse = false;
	bool swap = false;
	bool firstlast = false;
	int guide = 0;
	int gthresh = 30;
	bool agg = false;
	int nt = 100;
	bool post = true;
	int threshold = 15;
	int dthreshold = 9;
	bool blend = true;
	bool chroma = false;
	int y0 = 0;
	int y1 = 0;
	char ovr[255], *ovrp;
	bool show = false;
	bool debug = false;

	try
	{
		FILE *f;

		const char* plugin_dir = env->GetVar("$PluginDir$").AsString();
		strcpy(path, plugin_dir);
		strcat(path, "\\Telecide.def");
		if ((f = fopen(path, "r")) != NULL)
		{
			while(fgets(buf, 80, f) != 0)
			{
				if (strncmp(buf, "mm=", 3) == 0)
				{
					p = buf;
					while(*p++ != '=');
					mm = atoi(p);
				}
				if (strncmp(buf, "reverse=", 8) == 0)
				{
					p = buf;
					while(*p++ != '=');
					if (*p == 't') reverse = true;
					else reverse = false;
				}
				if (strncmp(buf, "swap=", 5) == 0)
				{
					p = buf;
					while(*p++ != '=');
					if (*p == 't') swap = true;
					else swap = false;
				}
				if (strncmp(buf, "firstlast=", 10) == 0)
				{
					p = buf;
					while(*p++ != '=');
					if (*p == 't') firstlast = true;
					else firstlast = false;
				}
				if (strncmp(buf, "guide=", 6) == 0)
				{
					p = buf;
					while(*p++ != '=');
					guide = atoi(p);
				}
				if (strncmp(buf, "gthresh=", 8) == 0)
				{
					p = buf;
					while(*p++ != '=');
					gthresh = atoi(p);
				}
				if (strncmp(buf, "agg=", 4) == 0)
				{
					p = buf;
					while(*p++ != '=');
					if (*p == 't') agg = true;
					else agg = false;
				}
				if (strncmp(buf, "nt=", 3) == 0)
				{
					p = buf;
					while(*p++ != '=');
					nt = atoi(p);
				}
				if (strncmp(buf, "post=", 5) == 0)
				{
					p = buf;
					while(*p++ != '=');
					if (*p == 't') post = true;
					else post = false;
				}
				if (strncmp(buf, "threshold=", 10) == 0)
				{
					p = buf;
					while(*p++ != '=');
					threshold = atoi(p);
				}
				if (strncmp(buf, "dthreshold=", 11) == 0)
				{
					p = buf;
					while(*p++ != '=');
					dthreshold = atoi(p);
				}
				if (strncmp(buf, "blend=", 6) == 0)
				{
					p = buf;
					while(*p++ != '=');
					if (*p == 't') blend = true;
					else blend = false;
				}
				if (strncmp(buf, "chroma=", 7) == 0)
				{
					p = buf;
					while(*p++ != '=');
					if (*p == 't') chroma = true;
					else chroma = false;
				}
				if (strncmp(buf, "y0=", 3) == 0)
				{
					p = buf;
					while(*p++ != '=');
					y0 = atoi(p);
				}
				if (strncmp(buf, "y1=", 3) == 0)
				{
					p = buf;
					while(*p++ != '=');
					y1 = atoi(p);
				}
				ovrp = ovr;
				*ovrp = 0;
				if (strncmp(buf, "ovr=", 4) == 0)
				{
					p = buf;
					while(*p++ != '"');
					while(*p != '\n' && *p != '\r' && *p != 0 && *p != EOF && *p != '"') *ovrp++ = *p++;
					*ovrp = 0;
				}
				if (strncmp(buf, "show=", 5) == 0)
				{
					p = buf;
					while(*p++ != '=');
					if (*p == 't') show = true;
					else show = false;
				}
				if (strncmp(buf, "debug=", 6) == 0)
				{
					p = buf;
					while(*p++ != '=');
					if (*p == 't') debug = true;
					else debug = false;
				}
			}
			fclose(f);
		}
	}
	catch (...)
	{
		// plugin directory not set
		// probably using an older version avisynth
	}

    return new Telecide(args[0].AsClip(),
						args[1].AsInt(mm),			// match mode
						args[2].AsBool(reverse),	// reverse
						args[3].AsBool(swap),	// swap
						args[4].AsBool(firstlast),	// firstlast
						args[5].AsInt(guide),		// guide
						args[6].AsInt(gthresh),		// gthresh
						args[7].AsBool(agg),	// aggressive pattern guidance
						args[8].AsInt(nt),		// noise threshold
						args[9].AsBool(post),	// postprocess
						args[10].AsInt(threshold),		// threshold
						args[11].AsInt(dthreshold),		// dthreshold
						args[12].AsBool(blend),	// blend
						args[13].AsBool(chroma),	// chroma
						args[14].AsInt(y0),		// y0
						args[15].AsInt(y1),		// y1
						args[16].AsString(ovr), // overrides file
						args[17].AsBool(show),	// overlay info on frames
						args[18].AsBool(debug),	// debug
						env);
}
